import asyncio
import functools
import os
import unittest
import inspect
from inspect import iscoroutine, iscoroutinefunction
from typing import Callable, Type, Coroutine, Union, Any

from danielutils import create_directory, get_current_working_directory, set_current_working_directory, delete_directory


def dispatch_function(func: Union[Callable, Coroutine], *args, **kwargs) -> Any:
    if asyncio.iscoroutinefunction(func):
        return asyncio.run(func(*args, **kwargs))
    elif asyncio.iscoroutine(func):
        return asyncio.run(func)
    else:
        return func(*args, **kwargs)


def improved_setup(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(self):
        self.cwd = f"./{self.__class__.__name__}_test_folder"
        create_directory(self.cwd)
        self.prev_cwd = get_current_working_directory()
        set_current_working_directory(os.path.join(self.prev_cwd, self.cwd))
        if func is not None:
            func(self)

    @functools.wraps(func)
    async def async_wrapper(self):
        self.cwd = f"./{self.__class__.__name__}_test_folder"
        create_directory(self.cwd)
        self.prev_cwd = get_current_working_directory()
        set_current_working_directory(os.path.join(self.prev_cwd, self.cwd))
        if func is not None:
            await func(self)

    return async_wrapper if iscoroutinefunction(func) else wrapper


def improved_teardown(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(self):
        if func is not None:
            func(self)
        set_current_working_directory(self.prev_cwd)
        delete_directory(self.cwd)

    @functools.wraps(func)
    async def async_wrapper(self):
        if func is not None:
            await func(self)
        set_current_working_directory(self.prev_cwd)
        delete_directory(self.cwd)

    return async_wrapper if iscoroutinefunction(func) else wrapper


class AutoCWDTestCase(unittest.TestCase):
    @staticmethod
    def _dummy(*args, **kwargs) -> None:
        pass

    @classmethod
    def __init_subclass__(cls: Type, **kwargs) -> None:
        dct = dict(cls.__dict__)
        impl_setUp = dct.get("setUp", cls._dummy)
        impl_tearDown = dct.get("tearDown", cls._dummy)
        setattr(cls, "setUp", improved_setup(impl_setUp))
        setattr(cls, "tearDown", improved_teardown(impl_tearDown))


class AsyncAutoCWDTestCase(unittest.IsolatedAsyncioTestCase):
    @staticmethod
    async def _async_dummy(*args, **kwargs) -> None:
        pass

    @classmethod
    def __init_subclass__(cls: Type, **kwargs) -> None:
        dct = dict(cls.__dict__)
        async_impl_setUp = dct.get("asyncSetUp", cls._async_dummy)
        async_impl_tearDown = dct.get("asyncTearDown", cls._async_dummy)
        setattr(cls, "asyncSetUp", improved_setup(async_impl_setUp))
        setattr(cls, "asyncTearDown", improved_teardown(async_impl_tearDown))


__all__ = [
    'AutoCWDTestCase',
    "AsyncAutoCWDTestCase"
]
